//GENERATED BY genfields.pl. DO NOT EDIT!
package vrml.external.field;
import vrml.external.*;
import java.util.*;

public class EventInSFBool extends EventIn {
		FreeWRLBrowser browser;
		String nodeid;
		String id;
		public EventInSFBool(FreeWRLBrowser b, String n, String i) {
			browser = b;
			nodeid = n;
			id = i;
			System.out.println("New SFBool: "+n+" "+id);
		}
		public void setValue(boolean value) 
			 {
				boolean v;
;
				;
				v = value;
				browser.send__eventin(nodeid,id, (v? "1": "0")) ;
		}
	}
	