// GENERATED BY genfields.pl. DO NOT EDIT!
package vrml.field;
import vrml.*;
import java.util.*;

public class ConstSFBool extends ConstField {
boolean v;
public ConstSFBool(boolean value) { v = value;}
public ConstSFBool(String s) throws Exception {
		;
		if(s == null) {
			v = false;; return;
		}
		s = s.trim();
		
   	s = s.trim();
	if(s.equals("1")) {v = true;}
	else if(s.equals("0") || s.equals("")) {v = false;}
	else {throw new Exception("Invalid boolean '"+s+"'");}

	}public boolean getValue() {return v;}
public String toString() {return (v? "1": "0");}public Object clone() {ConstSFBool _x = new ConstSFBool(v); return _x;}}