# AUTOMATICALLY GENERATED BY TOP-LEVEL MAKEFILE.PL -- DO NOT EDIT
%VRML::Config = (
	JPEG_DEFINE	=>	"",
	JS_INC	=>	"\-Ijs\/Linux_All_DBG\.OBJ\ \-Ijs\/",
	JS_LIBS	=>	"\-Ljs\/Linux_All_DBG\.OBJ\ \-ljs",
	FREETYPE_INC	=>	"",
	FREETYPE_LIBS	=>	"\-lttf",
	PNG_INC	=>	"",
	PNG_LIBS	=>	"\-lpng\ \-lz",
	FREETYPE_DEFINE	=>	"",
	OPENGL_INC	=>	"",
	OPENGL_LIBS	=>	"\-L\/BIG3\/Mesa\-3\.0\/lib\ \-L\/home\/l\/u\/lukka\/lib\ \-L\/usr\/X11R6\/lib\ \-L\/usr\/lib\/mesa\ \-lMesaGL_static\ \-lMesaGLU_static\ \-lXext\ \-lX11",
	PNG_DEFINE	=>	"",
	JPEG_INC	=>	"",
	JPEG_LIBS	=>	"\-ljpeg",
	OPENGL_DEFINE	=>	"",
VERSION => "0.14"
);
1;
