// GENERATED BY genfields.pl. DO NOT EDIT!
package vrml.field;
import vrml.*;
import java.util.*;

public class ConstSFTime extends ConstField {
double v;
public ConstSFTime(double val) { v=val;}
public ConstSFTime(String s) throws Exception {
		;
		if(s == null) {
			v = 0;; return;
		}
		s = s.trim();
		
	s = s.trim();
	v = new Double(s).doubleValue();

	}public double getValue() {return v;}
public String toString() {return new Double(v).toString();}public Object clone() {ConstSFTime _x = new ConstSFTime(v); return _x;}}