//GENERATED BY genfields.pl. DO NOT EDIT!
package vrml.external.field;
import vrml.external.*;
import java.util.*;

public class EventOutSFColor extends EventOut {
		float red,green,blue;

		public void value__set(String s) throws Exception {
			;
			if(s == null || s.equals("")) {
				red=0; green=0; blue=0;; return;
			}
			s = s.trim();
			
   	StringTokenizer tok = new StringTokenizer(s);
	red = 	new Float(tok.nextToken()).floatValue();
	green =	new Float(tok.nextToken()).floatValue();
	blue =	new Float(tok.nextToken()).floatValue();
	
		}
	public void getValue(float colors[]) {colors[0] = red; colors[1] = green; colors[2] = blue;}
public float getBlue() {return blue;}
public float getGreen() {return green;}
public float getRed() {return red;}

		};
	