// GENERATED BY genfields.pl. DO NOT EDIT!
package vrml.field;
import vrml.*;
import java.util.*;

public class ConstSFFloat extends ConstField {
float v;
public ConstSFFloat(float val) { v=val;}
public ConstSFFloat(String s) throws Exception {
		;
		if(s == null) {
			v = 0;; return;
		}
		s = s.trim();
		
	s = s.trim();
	v = new Float(s).floatValue();

	}public float getValue() {return v;}
public String toString() {return new Float(v).toString();}public Object clone() {ConstSFFloat _x = new ConstSFFloat(v); return _x;}}