//GENERATED BY genfields.pl. DO NOT EDIT!
package vrml.external.field;
import vrml.external.*;
import java.util.*;

public class EventOutSFBool extends EventOut {
		boolean v;

		public void value__set(String s) throws Exception {
			;
			if(s == null || s.equals("")) {
				v = false;; return;
			}
			s = s.trim();
			
   	s = s.trim();
	if(s.equals("1")) {v = true;}
	else if(s.equals("0") || s.equals("")) {v = false;}
	else {throw new Exception("Invalid boolean '"+s+"'");}

		}
	public boolean getValue() {return v;}

		};
	