// Copyright (C) 1998 Tuomas J. Lukka
// DISTRIBUTED WITH NO WARRANTY, EXPRESS OR IMPLIED.
// See the GNU Library General Public License (file COPYING in the distribution)
// for conditions of use and redistribution.

package vrml.external.field;

public class EventIn {
  // Get the type of this EventIn (specified in FieldTypes.java)
  public int           getType() { return 0; };
}
